# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""NSGA-II algorithm: https://ieeexplore.ieee.org/document/996017."""

from typing import Callable, Optional, Tuple

import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.evolution import numpy_populations
from vizier._src.algorithms.evolution import templates

Population = numpy_populations.Population
Offspring = numpy_populations.Offspring
Mutation = templates.Mutation


def _pareto_rank(ys: np.ndarray) -> np.ndarray:
  """Pareto rank, which is the number of points dominating it.

  Args:
    ys: (number of population) x (number of metrics) array.

  Returns:
    (number of population) integer array.
  """
  if ys.shape[0] == 0:
    return np.zeros([0])
  n = ys.shape[0]
  dominated = np.asarray(
      [
          [np.all(ys[i] <= ys[j]) & np.any(ys[j] > ys[i]) for i in range(n)]
          for j in range(n)
      ]
  )
  return np.sum(dominated, axis=0)


def _crowding_distance(ys: np.ndarray) -> np.ndarray:
  """Crowding distance.

  Args:
    ys: (number of population) x (number of metrics) array.

  Returns:
    (number of population) float32 array. Higher numbers mean less crowding
    and more desirable.
  """
  scores = np.zeros([ys.shape[0]], dtype=np.float32)
  for m in range(ys.shape[1]):
    # Sort by the m-th metric.
    yy = ys[:, m]  # Shape: (num_population,)
    sid = np.argsort(yy)

    # Boundary are assigned infinity.
    scores[sid[0]] += np.inf
    scores[sid[-1]] += np.inf

    # Compute the crowding distance.
    yrange = yy[sid[-1]] - yy[sid[0]] + np.finfo(np.float32).eps
    scores[sid[1:-1]] += (yy[sid[2:]] - yy[sid[:-2]]) / yrange
  return scores


def _constraint_violation(ys: np.ndarray) -> np.ndarray:
  """Counts the constraints violated.

  Args:
    ys: (number of population) x (number of metrics) array.

  Returns:
    (number of population) array of integers.
  """
  return np.sum(ys < 0, axis=1)


def _select_by(ys: np.ndarray, target: int) -> Tuple[np.ndarray, np.ndarray]:
  """Returns a boolean index array for the top `target` elements of ys.

  This method is tough to parse. Please improve the API if you see a better
  design!

  Args:
    ys: Array of shape [M]. Entries are expected to have a small set of unique
      values.
    target: Count to return.

  Returns:
    A tuple of two bolean index arrays `top` and `border`.
     * `ys[top]` has length less than or equal to `target`. They are within
       top `target`.
     * `ys[top | border]` has length greater than or equal to `target`.
     * `ys[border]` have all-identical entries. Callers should break ties
       among them.
     * `top & border` is all False.
  """
  if ys.shape[0] <= target:
    return (
        np.ones(ys.shape[:1], dtype=np.bool_),
        np.zeros(ys.shape[:1], dtype=np.bool_),
    )
  unique, counts = np.unique(ys, return_counts=True)
  cutoffidx = np.argmax(np.cumsum(counts) > target)
  cutoffnumber = unique[cutoffidx]
  return ys < cutoffnumber, ys == cutoffnumber


class NSGA2Survival(templates.Survival):
  """NSGA2 Survival step.

  Reference: https://ieeexplore.ieee.org/document/996017
  """

  def __init__(
      self,
      target_size: int,
      *,
      ranking_fn: Callable[[np.ndarray], np.ndarray] = _pareto_rank,
      eviction_limit: Optional[int] = None
  ):
    """Init.

    Args:
      target_size: select() method reduces the population to this size, unless
        the population is already at or below this size in which case select()
        is a no-op.
      ranking_fn: Takes (number of population) x (number of metrics) array of
        floating numbers and returns (number of population) array of integers,
        representing the pareto rank aka number of points it is dominated by.
        For performance-sensitive applications, plug in an XLA-compiled
        function.
      eviction_limit: Evict genes that were alive for this many generations.
    """
    self._target_size = target_size
    self._ranking_fn = ranking_fn
    self._eviction_limit = eviction_limit or float('inf')

  def select(self, population: Population) -> Population:
    """Applies survival process.

    Sorted all points by 3-tuple
    1. Descending order of safety constraint violation score. Zero
      means no violations.
    2. Ascending order of how many points it's dominated by.
    3. Descending order of crowding distance.

    Args:
      population:

    Returns:
      Population of size self._population_size.
    """
    population = population[population.ages < self._eviction_limit]
    if not population:
      # return empty.
      return population

    selected = population.empty_like()
    # Sort by the safety constraint.
    if selected.cs.shape[1]:
      top, border = _select_by(
          _constraint_violation(population.cs), target=self._target_size
      )
      selected += population[top]
      population = population[border]

    # Sort by the pareto rank.
    pareto_ranks = self._ranking_fn(population.ys)
    top, border = _select_by(
        pareto_ranks, target=self._target_size - len(selected)
    )
    selected += population[top]
    population = population[border]

    # Sort by the distance. Include the points that are already selected for
    # the computation.
    # Flip the sign so it works with ascending sort.
    distance = -_crowding_distance((selected + population).ys)
    sids = np.argsort(distance)
    # Selected points have fewer constraint violations or better pareto rank.
    # Regardless of the distance, they remain selected. Rank the remainder only.
    sids = sids[sids >= len(selected)] - len(selected)
    selected += population[sids[: self._target_size - len(selected)]]

    return attr.evolve(selected, ages=selected.ages + 1)


class NSGA2Designer(
    templates.CanonicalEvolutionDesigner[Population, Offspring]
):
  """NSGA2 Designer.

  Reference: https://ieeexplore.ieee.org/document/996017
  """

  def __init__(
      self,
      problem: vz.ProblemStatement,
      population_size: int = 50,
      first_survival_after: Optional[int] = None,
      *,
      ranking_fn: Callable[[np.ndarray], np.ndarray] = _pareto_rank,
      eviction_limit: Optional[int] = None,
      adaptation: Optional[Mutation[Population, Offspring]] = None,
      adaptation_callable: Optional[
          Callable[[int], Mutation[Population, Offspring]]
      ] = None,
      metadata_namespace: str = 'nsga2',
      seed: Optional[int] = None
  ):
    """Creates NSGA2 Designer.

    Args:
      problem:
      population_size: Survival steps reduce the population to this size.
      first_survival_after: Apply the survival step after observing this many
        trials. Leave it unset to use the default behavior.
      ranking_fn: Takes (number of population) x (number of metrics) array of
        floating numbers and returns (number of population) array of integers,
        representing the pareto rank aka number of points it is dominated by.
        The default implementation is reasonably fast for hundreds of trials,
        but if you want to improve performance, your own implementation can be
        injected.
      eviction_limit: Evict a gene that has been alive for this many
        generations.
      adaptation: Fixed mutation used to evolve population.
      adaptation_callable: If specified, adaptation callable is a function of
        num_trials that returns the trial-dependent adaptation.
      metadata_namespace: Metadata namespace to use.
      seed: Random seed.

    Returns:
      NSGA2 Designer.
    """
    super().__init__(
        numpy_populations.PopulationConverter(
            problem.search_space,
            problem.metric_information,
            metadata_ns=metadata_namespace,
        ),
        numpy_populations.UniformRandomSampler(problem.search_space, seed=seed),
        NSGA2Survival(
            population_size,
            ranking_fn=ranking_fn,
            eviction_limit=eviction_limit,
        ),
        adaptation=adaptation
        or numpy_populations.LinfMutation(seed=seed, norm=0.001),
        first_survival_after=first_survival_after,
        adaptation_callable=adaptation_callable,
        population_size=population_size,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for proto_converters."""

from absl import logging

import attr

from vizier._src.pyvizier.oss import proto_converters
from vizier._src.pyvizier.pythia import study
from vizier._src.pyvizier.shared import parameter_config as pc
from vizier._src.pyvizier.shared import trial
from vizier._src.service import study_pb2

from google.protobuf import struct_pb2
from google.protobuf import wrappers_pb2
from vizier._src.pyvizier.oss import compare
from absl.testing import absltest
from absl.testing import parameterized

Metric = trial.Metric
Measurement = trial.Measurement


class StudyStateConverterTest(absltest.TestCase):

  def testReversibility(self):
    for original_pyvizier_state in study.StudyState:
      proto_state = proto_converters.StudyStateConverter.to_proto(
          original_pyvizier_state
      )
      py_state = proto_converters.StudyStateConverter.from_proto(proto_state)
      self.assertEqual(original_pyvizier_state, py_state)

  def testStudyStateNotSet(self):
    self.assertEqual(
        proto_converters.StudyStateConverter.from_proto(
            study_pb2.Study.State.STATE_UNSPECIFIED
        ),
        study.StudyState.ACTIVE,
    )


class MeasurementConverterTest(absltest.TestCase):

  def testMeasurementProtoWithEmptyNamedMetric(self):
    proto = study_pb2.Measurement()
    proto.metrics.add(metric_id='', value=0.8)
    measurement = proto_converters.MeasurementConverter.from_proto(proto)
    self.assertEqual(measurement.metrics[''], Metric(value=0.8))

  def testMeasurementCreation(self):
    measurement = Measurement(
        metrics={
            '': Metric(
                value=0
            ),  # The empty metric always exists in Measurement.
            'pr-auc': Metric(value=0.8),
            'latency': Metric(value=32),
        },
        elapsed_secs=12,
        steps=12,
    )
    proto = proto_converters.MeasurementConverter.to_proto(measurement)
    self.assertEqual(
        attr.asdict(proto_converters.MeasurementConverter.from_proto(proto)),
        attr.asdict(measurement),
    )


ParameterValue = trial.ParameterValue


class ParameterValueConverterTest(parameterized.TestCase):

  def testToDoubleProto(self):
    value = ParameterValue(True)
    compare.assertProto2Equal(
        self,
        proto_converters.ParameterValueConverter.to_proto(value, 'aa'),
        study_pb2.Trial.Parameter(
            parameter_id='aa', value=struct_pb2.Value(number_value=1.0)
        ),
    )

  def testToDiscreteProto(self):
    value = ParameterValue(True)
    compare.assertProto2Equal(
        self,
        proto_converters.ParameterValueConverter.to_proto(value, 'aa'),
        study_pb2.Trial.Parameter(
            parameter_id='aa', value=struct_pb2.Value(number_value=1.0)
        ),
    )

  def testToStringProto(self):
    value = ParameterValue('category')
    compare.assertProto2Equal(
        self,
        proto_converters.ParameterValueConverter.to_proto(value, 'aa'),
        study_pb2.Trial.Parameter(
            parameter_id='aa', value=struct_pb2.Value(string_value='category')
        ),
    )

  def testToIntegerProto(self):
    value = ParameterValue(True)
    compare.assertProto2Equal(
        self,
        proto_converters.ParameterValueConverter.to_proto(value, 'aa'),
        study_pb2.Trial.Parameter(
            parameter_id='aa', value=struct_pb2.Value(number_value=1.0)
        ),
    )


class TrialConverterTest(absltest.TestCase):

  def testFromProtoCompleted(self):
    proto = study_pb2.Trial(id=str(1))
    proto.state = study_pb2.Trial.State.SUCCEEDED
    proto.parameters.add(
        parameter_id='float', value=struct_pb2.Value(number_value=1.0)
    )
    proto.parameters.add(
        parameter_id='int', value=struct_pb2.Value(number_value=2)
    )
    proto.parameters.add(
        parameter_id='str', value=struct_pb2.Value(string_value='3')
    )
    proto.final_measurement.metrics.add(metric_id='pr-auc', value=0.8)
    proto.final_measurement.metrics.add(metric_id='latency', value=32)

    proto.start_time.seconds = 1586649600
    proto.end_time.seconds = 1586649600 + 10

    proto.measurements.add(step_count=10)
    proto.measurements[-1].elapsed_duration.seconds = 15
    proto.measurements[-1].metrics.add(metric_id='pr-auc', value=0.7)
    proto.measurements[-1].metrics.add(metric_id='latency', value=42)

    proto.measurements.add(step_count=20)
    proto.measurements[-1].elapsed_duration.seconds = 30
    proto.measurements[-1].metrics.add(metric_id='pr-auc', value=0.75)
    proto.measurements[-1].metrics.add(metric_id='latency', value=37)

    test = proto_converters.TrialConverter.from_proto(proto=proto)
    self.assertEqual(test.id, 1)
    self.assertEqual(test.status, trial.TrialStatus.COMPLETED)
    self.assertTrue(test.is_completed)
    self.assertFalse(test.infeasible)
    self.assertIsNone(test.infeasibility_reason)
    self.assertLen(test.parameters, 3)
    self.assertEqual(test.parameters['float'].value, 1.0)
    self.assertEqual(test.parameters['int'].value, 2)
    self.assertEqual(test.parameters['str'].value, '3')

    # Final measurement
    self.assertLen(test.final_measurement.metrics, 2)
    self.assertEqual(test.final_measurement.metrics['pr-auc'].value, 0.8)
    self.assertEqual(test.final_measurement.metrics['latency'].value, 32)

    # Intermediate measurement
    self.assertEqual(
        test.measurements[0],
        trial.Measurement(
            metrics={'pr-auc': 0.7, 'latency': 42}, steps=10, elapsed_secs=15
        ),
    )
    self.assertEqual(
        test.measurements[1],
        trial.Measurement(
            metrics={'pr-auc': 0.75, 'latency': 37}, steps=20, elapsed_secs=30
        ),
    )

    self.assertEqual(test.id, 1)

    self.assertIsNotNone(test.creation_time)
    self.assertIsNotNone(test.completion_time)
    self.assertEqual(test.duration.total_seconds(), 10)

    self.assertFalse(test.infeasible)

  def testFromProtoPending(self):
    proto = study_pb2.Trial(id=str(2))
    proto.state = study_pb2.Trial.State.ACTIVE
    proto.start_time.seconds = 1586649600
    test = proto_converters.TrialConverter.from_proto(proto=proto)
    self.assertEqual(test.status, trial.TrialStatus.ACTIVE)
    self.assertFalse(test.is_completed)
    self.assertFalse(test.infeasible)
    self.assertIsNone(test.infeasibility_reason)
    self.assertIsNotNone(test.creation_time)
    self.assertIsNone(test.completion_time)
    self.assertIsNone(test.duration)
    self.assertEmpty(test.metadata)

  def testFromProtoInfeasible(self):
    proto = study_pb2.Trial(id=str(1))
    proto.state = study_pb2.Trial.State.INFEASIBLE
    proto.parameters.add(
        parameter_id='float', value=struct_pb2.Value(number_value=1.0)
    )
    proto.parameters.add(
        parameter_id='int', value=struct_pb2.Value(number_value=2)
    )
    proto.parameters.add(
        parameter_id='str', value=struct_pb2.Value(string_value='3')
    )
    proto.start_time.seconds = 1586649600
    proto.end_time.seconds = 1586649600 + 10
    proto.infeasible_reason = 'A reason'

    test = proto_converters.TrialConverter.from_proto(proto=proto)
    self.assertEqual(test.status, trial.TrialStatus.COMPLETED)
    self.assertTrue(test.is_completed)
    self.assertTrue(test.infeasible)
    self.assertEqual(test.infeasibility_reason, 'A reason')

  def testFromProtoInvalidTrial(self):
    proto = study_pb2.Trial(id=str(2))
    proto.parameters.add(
        parameter_id='float', value=struct_pb2.Value(number_value=1.0)
    )
    proto.parameters.add(
        parameter_id='float', value=struct_pb2.Value(number_value=2.0)
    )
    proto.state = study_pb2.Trial.State.ACTIVE
    proto.start_time.seconds = 1586649600
    with self.assertRaisesRegex(ValueError, 'Invalid trial proto'):
      proto_converters.TrialConverter.from_proto(proto=proto)

  def testFromProtoMetadata(self):
    proto = study_pb2.Trial(id=str(1))
    proto.state = study_pb2.Trial.ACTIVE
    proto.parameters.add(
        parameter_id='float', value=struct_pb2.Value(number_value=1.0)
    )
    proto.metadata.add(key='key0', ns='x', value='namespace=x0')
    proto.metadata.add(key='key1', ns='x', value='namespace=x1')
    proto.metadata.add(key='key1', ns='', value='gets overwritten')
    proto.metadata.add(key='key1', value='second value takes priority')
    logging.info('PROTO:: %s', proto)
    added1 = proto.metadata.add(key='proto')
    added1.proto.Pack(study_pb2.Trial(id=str(999)))
    added2 = proto.metadata.add(key='proto', ns='t')
    added2.proto.Pack(study_pb2.Trial(id=str(991)))
    test = proto_converters.TrialConverter.from_proto(proto=proto)
    logging.info('TEST:: %s', repr(test.metadata))
    logging.info('TEST-x:: %s', repr(test.metadata.ns('x')))
    logging.info('TEST-t:: %s', repr(test.metadata.ns('t')))
    logging.info('TEST-namespaces: %s', test.metadata.namespaces())
    logging.info('test.ns("x"):: %s', test.metadata.ns('x'))
    self.assertEqual(test.metadata['key1'], 'second value takes priority')
    self.assertEqual(test.metadata.abs_ns(['x'])['key0'], 'namespace=x0')
    self.assertEqual(test.metadata.abs_ns(['x'])['key1'], 'namespace=x1')
    self.assertEqual(
        test.metadata.get_proto('proto', cls=study_pb2.Trial),
        study_pb2.Trial(id=str(999)),
    )
    self.assertEqual(
        test.metadata.abs_ns(['t']).get_proto('proto', cls=study_pb2.Trial),
        study_pb2.Trial(id=str(991)),
    )


class TrialConverterToProtoTest(absltest.TestCase):
  """Tests for TrialConverter.to_proto()."""

  def _GetSingleObjectiveBaseTrial(self):
    proto = study_pb2.Trial(
        name='owners/my_username/studies/cifar_10',
        id=str(2),
        client_id='worker0',
    )
    proto.parameters.add(
        parameter_id='activation', value=struct_pb2.Value(string_value='relu')
    )
    proto.parameters.add(
        parameter_id='synchronus', value=struct_pb2.Value(string_value='true')
    )
    proto.parameters.add(
        parameter_id='batch_size', value=struct_pb2.Value(number_value=32)
    )
    proto.parameters.add(
        parameter_id='floating_point_param',
        value=struct_pb2.Value(number_value=32.0),
    )
    proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5)
    )
    proto.parameters.add(
        parameter_id='units', value=struct_pb2.Value(number_value=50)
    )
    proto.start_time.seconds = 1630505100
    return proto

  def testParameterBackToBackConversion(self):
    proto = self._GetSingleObjectiveBaseTrial()
    proto.state = study_pb2.Trial.State.ACTIVE
    pytrial = proto_converters.TrialConverter.from_proto(proto)
    got = proto_converters.TrialConverter.to_proto(pytrial)
    compare.assertProto2Equal(self, proto, got)

  def testFinalMeasurementBackToBackConversion(self):
    proto = study_pb2.Trial(id=str(1), state=study_pb2.Trial.State.SUCCEEDED)
    proto.start_time.seconds = 12456
    proto.end_time.seconds = 12456 + 10
    proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5)
    )
    proto.final_measurement.step_count = 101
    proto.final_measurement.elapsed_duration.seconds = 67
    proto.final_measurement.metrics.add(metric_id='loss', value=56.8)
    proto.final_measurement.metrics.add(metric_id='objective', value=77.7)
    proto.final_measurement.metrics.add(metric_id='objective2', value=-0.2)

    pytrial = proto_converters.TrialConverter.from_proto(proto)
    got = proto_converters.TrialConverter.to_proto(pytrial)
    compare.assertProto2SameElements(self, proto, got, number_matters=True)

  def testMeasurementBackToBackConversion(self):
    proto = study_pb2.Trial(
        id=str(2), state=study_pb2.Trial.State.ACTIVE, client_id='worker0'
    )
    proto.start_time.seconds = 1630505100
    proto.measurements.add(step_count=123)
    proto.measurements[-1].elapsed_duration.seconds = 22
    proto.measurements[-1].metrics.add(metric_id='objective', value=0.4321)
    proto.measurements[-1].metrics.add(metric_id='loss', value=0.001)

    proto.measurements.add(step_count=789)
    proto.measurements[-1].elapsed_duration.seconds = 55
    proto.measurements[-1].metrics.add(metric_id='objective', value=0.21)
    proto.measurements[-1].metrics.add(metric_id='loss', value=0.0001)

    pytrial = proto_converters.TrialConverter.from_proto(proto)
    got = proto_converters.TrialConverter.to_proto(pytrial)
    compare.assertProto2SameElements(self, proto, got, number_matters=True)


class ParameterConfigConverterToProtoTest(absltest.TestCase):
  """Tests for ParameterConfigConverter.to_proto()."""

  def testDiscreteConfigToProto(self):
    feasible_values = (-1, 3, 2)
    parameter_config = pc.ParameterConfig.factory(
        'name',
        feasible_values=feasible_values,
        scale_type=pc.ScaleType.LOG,
        default_value=2,
    )

    proto = proto_converters.ParameterConfigConverter.to_proto(parameter_config)
    self.assertEqual(proto.parameter_id, 'name')
    self.assertEqual(proto.discrete_value_spec.values, [-1.0, 2.0, 3.0])
    self.assertEqual(proto.discrete_value_spec.default_value.value, 2)
    self.assertEqual(
        proto.scale_type,
        study_pb2.StudySpec.ParameterSpec.ScaleType.UNIT_LOG_SCALE,
    )


class ParameterConfigConverterFromProtoTest(absltest.TestCase):
  """Tests for ParameterConfigConverter.from_proto()."""

  def testCreatesFromGoodProto(self):
    proto = study_pb2.StudySpec.ParameterSpec(
        parameter_id='name',
        discrete_value_spec=study_pb2.StudySpec.ParameterSpec.DiscreteValueSpec(
            values=[1.0, 2.0, 3.0],
            default_value=wrappers_pb2.DoubleValue(value=2.0),
        ),
    )
    parameter_config = proto_converters.ParameterConfigConverter.from_proto(
        proto
    )
    self.assertEqual(parameter_config.name, proto.parameter_id)
    self.assertEqual(parameter_config.type, pc.ParameterType.DISCRETE)
    self.assertEqual(parameter_config.bounds, (1.0, 3.0))
    self.assertEqual(parameter_config.feasible_values, [1.0, 2.0, 3.0])
    self.assertEqual(parameter_config.default_value, 2.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Benchmark runners to run benchmarks via modularized subroutines.

Modularization provides flexibility and clarity. Also, this allows easy
configuration of the protocol routine, which then can be applied to multiple
benchmarks via runner.run().

Ex: Typical Suggest + Evaluate loop all repeated for 7 iterations.

  # Define protocol (i.e. what should the benchmark do?).
  runner = benchmark_runner.BenchmarkRunner(
          benchmark_subroutines=[
              benchmark_runner.GenerateSuggestions(),
              benchmark_runner.EvaluateActiveTrials()
          ],
          num_repeats=7)

  # Initialize state (i.e. what problem and algorithm are we using?).
  benchmark_state =
  benchmark_runner.benchmark_state.BenchmarkState.from_designer_factory(
          designer_factory=designer_factory, experimenter=experimenter)

  # Runs benchmark protocol(s) and updates state.
  runner.run(benchmark_state)

Ex2: The power of modularization allows for complicated subroutines, such as
one that simulates the stochasticity of natural evaluation processes.

  # Allow for fixed-size batched Suggests and random-size evaluations.
  subroutines = []
  for _ in range(10):
    subroutines.append(benchmark_runner.GenerateSuggestions(num_suggestions=3)
    subroutines.append(benchmark_runner.EvaluateActiveTrial(
        num_trials_to_evaluate = np.randomInt(0, 10))
  runner = benchmark_runner.BenchmarkRunner(benchmark_subroutines=subroutines)
"""

import abc
import time
from typing import Optional, Sequence

from absl import logging
import attr
from vizier import pyvizier as vz
from vizier._src.benchmarks.runners import benchmark_state


class BenchmarkSubroutine(abc.ABC):
  """Abstraction for core benchmark routines.

  Benchmark protocols are modular alterations of BenchmarkState by reference.
  """

  @abc.abstractmethod
  def run(self, state: benchmark_state.BenchmarkState) -> None:
    """Abstraction to alter BenchmarkState by reference."""


@attr.define
class GenerateAndEvaluate(BenchmarkSubroutine):
  """Generate a fixed number of Suggestions and Evaluate them immediately.

  Use this Subroutine for simple benchmarks as it is computationally efficient.
  For benchmarks with complicated evaluation procedures, use a combination
  of other Subroutines.
  """

  # Number of total suggestions as a batch.
  batch_size: int = attr.field(
      default=1, validator=attr.validators.instance_of(int)
  )

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    suggestions = state.algorithm.suggest(self.batch_size)
    if not suggestions:
      logging.info(
          (
              'Algorithm did not generate %d suggestions'
              'because it returned nothing.'
          ),
          self.batch_size,
      )
    state.experimenter.evaluate(list(suggestions))


@attr.define
class GenerateSuggestions(BenchmarkSubroutine):
  """Generate a fixed number of Suggestions as Active Trials."""

  # Number of total suggestions as a batch.
  batch_size: int = attr.field(
      default=1, validator=attr.validators.instance_of(int)
  )

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    suggestions = state.algorithm.suggest(self.batch_size)
    if not suggestions:
      logging.info(
          (
              'Suggestions did not generate %d suggestions'
              'because designer returned nothing.'
          ),
          self.batch_size,
      )


@attr.define
class FillActiveTrials(BenchmarkSubroutine):
  """Generate a number of Suggestions up to a limit of Active Trials."""

  # Upper limit of active Trials to be filled with new Suggestions.
  num_active_trials_limit: int = attr.field(
      default=1, validator=attr.validators.instance_of(int)
  )

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    active_trials = state.algorithm.supporter.GetTrials(
        status_matches=vz.TrialStatus.ACTIVE
    )
    if (
        num_suggestions := self.num_active_trials_limit - len(active_trials)
    ) > 0:
      suggestions = state.algorithm.suggest(num_suggestions)
      if not suggestions:
        logging.info(
            (
                'Suggestions did not generate %d suggestions'
                'to fill active Trials to %d '
                'because designer returned nothing.'
            ),
            num_suggestions,
            self.num_active_trials_limit,
        )


@attr.define
class EvaluateActiveTrials(BenchmarkSubroutine):
  """Evaluate a fixed number of Active Trials as Completed Trials."""

  # If None, this Evaluates all Active Trials.
  num_evaluations: Optional[int] = attr.field(default=None)

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    active_trials = state.algorithm.supporter.GetTrials(
        status_matches=vz.TrialStatus.ACTIVE
    )
    if self.num_evaluations is None:
      evaluated_trials = active_trials
    else:
      evaluated_trials = active_trials[: self.num_evaluations]

    state.experimenter.evaluate(evaluated_trials)


@attr.define
class EvaluateAndAddPriorStudy(BenchmarkSubroutine):
  """Evaluate a fixed number of Active Trials as Completed Trials."""

  # Runner to use to mutate study.
  benchmark_runner: BenchmarkSubroutine = attr.field(
      kw_only=True,
      validator=attr.validators.instance_of(BenchmarkSubroutine),
  )
  # State factory.
  benchmark_state_factory: benchmark_state.BenchmarkStateFactory = attr.field(
      kw_only=True,
      validator=attr.validators.instance_of(
          benchmark_state.BenchmarkStateFactory
      ),
  )
  study_guid: Optional[str] = attr.field(
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )
  seed: Optional[int] = attr.field(
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(int)),
  )

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    prior_state = self.benchmark_state_factory(seed=self.seed)
    self.benchmark_runner.run(prior_state)
    prior_problem = prior_state.algorithm.supporter.GetStudyConfig()
    prior_trials = prior_state.algorithm.supporter.GetTrials()
    prior_study = vz.ProblemAndTrials(
        problem=prior_problem, trials=prior_trials
    )

    state.algorithm.supporter.SetPriorStudy(
        study=prior_study, study_guid=self.study_guid
    )


@attr.define
class BenchmarkRunner(BenchmarkSubroutine):
  """Run a sequence of subroutines, all repeated for a few iterations."""

  # A sequence of benchmark subroutines that alter BenchmarkState.
  benchmark_subroutines: Sequence[BenchmarkSubroutine] = attr.field()
  # Number of times to repeat applying benchmark_subroutines.
  num_repeats: int = attr.field(
      default=1, validator=attr.validators.instance_of(int)
  )

  def run(self, state: benchmark_state.BenchmarkState) -> None:
    """Run algorithm with benchmark subroutines with repetitions."""
    for iteration in range(self.num_repeats):
      start = time.time()
      for subroutine in self.benchmark_subroutines:
        subroutine.run(state)
      total_time = time.time() - start
      logging.info(
          'Completed BenchmarkRunner iteration %d out of %d (%f seconds)',
          iteration,
          self.num_repeats,
          total_time,
      )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scheduled GP-Bandit for budget-aware optimization.
"""

import random
from typing import Sequence
from absl import logging
import attrs
import jax
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_bandit
from vizier._src.algorithms.designers.gp import acquisitions


@attrs.define(auto_attribs=False)
class VizierScheduledGPBandit(vza.Designer):
  """Scheduled GP-Bandit.

  Attributes:
    problem: Must be a flat study with a single metric.
    max_num_trials: The designer budget in trials.
    rng: If not set, uses random numbers.
    init_ucb_coef: The initial UCB coefficient.
    final_ucb_coef: The final UCB coefficient.
    decay_rate: The decay rate.
    metadata_ns: Metadata namespace that this designer writes to.
  """

  _problem: vz.ProblemStatement = attrs.field(kw_only=False)
  _max_num_trials: int = attrs.field(
      kw_only=True, validator=attrs.validators.ge(0)
  )
  _rng: jax.Array = attrs.field(
      factory=lambda: jax.random.PRNGKey(random.getrandbits(32)), kw_only=True
  )
  _init_ucb_coef: float = attrs.field(default=3.26359, kw_only=True)
  _final_ucb_coef: float = attrs.field(default=0.01577, kw_only=True)
  _decay_rate: float = attrs.field(
      default=1.02390, kw_only=True, validator=attrs.validators.ge(0)
  )
  # ------------------------------------------------------------------
  # Internal attributes which should not be set by callers.
  # ------------------------------------------------------------------
  _gp_bandit_designer: gp_bandit.VizierGPBandit = attrs.field(init=False)
  _trials_count: int = attrs.field(init=False, default=0)
  _metadata_ns: str = attrs.field(
      default="oss_scheduled_gp_bandit", kw_only=True, init=False
  )

  def __attrs_post_init__(self):
    self._gp_bandit_designer = gp_bandit.VizierGPBandit(
        self._problem, rng=self._rng
    )

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the designer based on completed and pending trials."""
    self._trials_count += len(completed.trials)
    self._gp_bandit_designer.update(completed, all_active)

  def _compute_ucb_coefficient(self):
    """Compute the UCB coefficient."""
    alpha = 1 / self._decay_rate
    a = self._init_ucb_coef
    b = -np.log(self._final_ucb_coef / self._init_ucb_coef) / (
        (self._max_num_trials - 1) ** alpha
    )
    return a * np.exp(-b * self._trials_count**alpha)

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggest trials."""
    ucb_coef = self._compute_ucb_coefficient()
    logging.info(
        "UCB coefficient (trials count = %s): %s", self._trials_count, ucb_coef
    )
    self._gp_bandit_designer._scoring_function_factory = (  # pylint: disable=protected-access
        acquisitions.bayesian_scoring_function_factory(
            lambda _: acquisitions.UCB(ucb_coef)
        )
    )
    suggest_trials = self._gp_bandit_designer.suggest(count)
    for trial in suggest_trials:
      metadata = trial.metadata.ns(self._metadata_ns).ns("devinfo")
      metadata["ucb_coefficient"] = str(ucb_coef)
    return suggest_trials



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Ensembling design algorithms for choosing between algorithms/experts."""

import abc
import attrs
import numpy as np

# An observation, that is a expert index (int) along with its reward.
IndexWithReward = tuple[int, float]


class EnsembleDesign(abc.ABC):
  """Strategy that samples which index/expert to choose according to rewards."""

  @property
  @abc.abstractmethod
  def ensemble_probs(self) -> np.ndarray:
    """Computes the probability distribution used to choose the expert.

    The returned shape is (num_experts,) and is a probability distribution.
    """
    pass

  @abc.abstractmethod
  def update(self, observation: IndexWithReward, **kwargs):
    """Updates the strategy with observation."""
    pass


@attrs.define
class RandomEnsembleDesign(EnsembleDesign):
  indices: list[int] = attrs.field()

  @property
  def ensemble_probs(self) -> np.ndarray:
    """The probability distribution used to choose the expert."""
    return np.ones(shape=len(self.indices)) / len(self.indices)

  def update(self, observation: IndexWithReward, **kwargs):
    """Updates the strategy with observation."""
    pass


def softmax(x: np.ndarray) -> np.ndarray:
  """Compute softmax values for x."""
  e_x = np.exp(x - np.max(x))
  return e_x / np.sum(e_x)


# https://bjpcjp.github.io/pdfs/math/bandits-exp3-IX-BA.pdf
@attrs.define
class EXP3IXEnsembleDesign(EnsembleDesign):
  """The EXP3-IX Algorithm that is robust against small probabilities."""

  indices: list[int] = attrs.field()
  stepsize: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  max_reward: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )

  def __attrs_post_init__(self):
    self._log_weights = np.zeros(shape=len(self.indices))
    self._history: list[IndexWithReward] = []

  @property
  def ensemble_probs(self) -> np.ndarray:
    return softmax(self._log_weights)

  def update(self, observation: IndexWithReward):
    expert_idx, reward = observation
    reward = min(self.max_reward, reward)
    gamma = 1.0 / np.sqrt(1.0 + len(self._history))
    loss_estimator = (self.max_reward - reward) / (
        self.ensemble_probs[expert_idx] + gamma
    )

    self._log_weights[expert_idx] += self.stepsize * (-loss_estimator)
    self._history.append(observation)


# pytype: disable=attribute-error
# https://www.cs.princeton.edu/courses/archive/fall16/cos402/lectures/402-lec22.pdf.
@attrs.define
class EXP3UniformEnsembleDesign(EnsembleDesign):
  """The EXP3 algorithm with uniform exploration."""

  indices: list[int] = attrs.field()
  stepsize: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  max_reward: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  use_loss_formulation: bool = attrs.field(default=False)
  # If the fed-in rewards are estimators, no reward clipping or
  # importance weighting logic are done.
  use_reward_estimator: bool = attrs.field(default=False)

  def __attrs_post_init__(self):
    self._log_weights = np.zeros(shape=len(self.indices))
    self._history: list[IndexWithReward] = []

  @property
  def ensemble_probs(self) -> np.ndarray:
    n_arms = len(self.indices)
    uniform = np.ones(n_arms) * 1.0 / n_arms
    gamma = 1.0 / np.sqrt(1 + len(self._history))
    probs = (1 - gamma) * softmax(self._log_weights) + gamma * uniform
    return probs

  def update(self, observation: IndexWithReward):
    """Update history and weights."""
    expert_idx, reward = observation
    if not self.use_reward_estimator:
      reward = min(self.max_reward, reward)

    gamma = 1.0 / np.sqrt(1 + len(self._history))
    if self.use_loss_formulation:
      if self.use_reward_estimator:
        loss_estimator = self.max_reward - reward
      else:
        loss_estimator = (self.max_reward - reward) / (
            self.ensemble_probs[expert_idx]
        )
      self._log_weights[expert_idx] += -self.stepsize * gamma * loss_estimator
    else:
      if self.use_reward_estimator:
        reward_estimator = reward
      else:
        reward_estimator = reward / (self.ensemble_probs[expert_idx])
      self._log_weights[expert_idx] += self.stepsize * gamma * reward_estimator

    self._history.append(observation)

  @property
  def history(self) -> list[IndexWithReward]:
    return self._history


@attrs.define(kw_only=True)
class AdaptiveEnsembleDesign(EnsembleDesign):
  """An EnsembleStrategy that minimizes adaptive regret."""

  indices: list[int] = attrs.field()
  # List of max lengths
  max_lengths: list[int] = attrs.field()
  # Base stepsize and meta_stepsize should theoretically be 1/sqrt(n) where
  # n = # of indices (or underlying arms) when using a reward estimator.
  base_stepsize: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  meta_stepsize: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  max_reward: float = attrs.field(
      default=1.0,
      validator=[attrs.validators.instance_of(float), attrs.validators.gt(0)],
  )
  naive_sampling: bool = attrs.field(default=False)

  def __attrs_post_init__(self):
    self._log_weights = {}
    self._base_algos = {}
    self._history = []
    for max_length in self.max_lengths:
      # Initialize log_weight = log(1/sqrt(max_length * len(indices))).
      self._base_algos[max_length] = EXP3UniformEnsembleDesign(
          indices=self.indices,
          stepsize=self.base_stepsize,
          use_loss_formulation=False,
          use_reward_estimator=True,
      )
      self._log_weights[max_length] = (
          -np.log(max_length * len(self.indices)) / 2.0
      )

  @property
  def ensemble_probs(self) -> np.ndarray:
    # Return observation probabilities every other observation.
    if len(self._history) % 2 == 1:
      return self.observation_probs

    return self.play_probs

  @property
  def play_probs(self) -> np.ndarray:
    """Returns the probability of playing an expert."""
    ensemble_probs = [algo.ensemble_probs for algo in self._base_algos.values()]
    weights = softmax(np.array(list(self._log_weights.values())))
    final_probs = np.zeros(len(self.indices))
    for weight, probs in zip(weights, ensemble_probs):
      final_probs += weight * probs

    return final_probs

  @property
  def observation_probs(self) -> np.ndarray:
    """Returns the probability of observing an expert."""
    num_experts = len(self.indices)
    uniform = np.ones(num_experts) * 1.0 / num_experts
    if self.naive_sampling:
      return uniform

    algo_prob_max = np.zeros(num_experts)
    algo_prob_sum = np.zeros(num_experts)
    ensemble_probs = [algo.ensemble_probs for algo in self._base_algos.values()]
    for i in range(num_experts):
      algo_prob_max[i] = np.amax(np.array([p[i] ** 2 for p in ensemble_probs]))
      algo_prob_sum[i] = np.sum(np.array([p[i] for p in ensemble_probs]))

    # Take the average of the two max and sum distributions.
    observation_prob = algo_prob_max / (
        2 * np.sum(algo_prob_max)
    ) + algo_prob_sum / (2 * np.sum(algo_prob_sum))
    return observation_prob

  def update(self, observation: IndexWithReward):
    expert_idx, reward = observation
    reward = min(reward, self.max_reward)
    reward_estimator = reward * 1.0 / self.ensemble_probs[expert_idx]

    # Updates the states of base algorithms.
    for max_length, base_algo in self._base_algos.items():
      # If the history is about to be filled, restart.
      if len(base_algo.history) + 1 >= max_length:
        # Initialize log_weight = log(1/sqrt(max_length * len(indices))).
        self._log_weights[max_length] = (
            -np.log(max_length * len(self.indices)) / 2.0
        )
        self._base_algos[max_length] = EXP3UniformEnsembleDesign(
            indices=self.indices,
            stepsize=self.base_stepsize,
            use_loss_formulation=False,
            use_reward_estimator=True,
        )

        continue

      # Otherwise, update meta-weights via unbiased reward estimation.
      reward_estimator_base = reward_estimator * (
          base_algo.ensemble_probs[expert_idx] - self.play_probs[expert_idx]
      )
      # Stepsize is 1/sqrt(max_length).
      gamma = 1.0 / np.sqrt(max_length)
      self._log_weights[max_length] += (
          self.meta_stepsize * gamma * reward_estimator_base
      )

      # Propagate observed reward estimator.
      base_algo.update((expert_idx, reward_estimator))

    self._history.append(observation)


# pytype: enable=attribute-error



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Flax module for a trainable stochastic process."""

import abc
from typing import Any, Callable, Generator, Generic, Optional, Protocol, TypeVar

from absl import logging
import attr
import chex
import equinox as eqx
from flax import config as flax_config
from flax import linen as nn
from flax import struct
import jax
from jax import numpy as jnp
from jax import tree_util
from jax.typing import ArrayLike
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import types

flax_config.update('flax_return_frozendict', False)

tfd = tfp.distributions
tfb = tfp.bijectors
tfde = tfp.experimental.distributions
tfpke = tfp.experimental.psd_kernels

_D = TypeVar('_D', bound=tfd.Distribution)


class InitFn(Protocol):
  """Protocol for Flax parameter initialization functions."""

  @abc.abstractmethod
  def __call__(self, rng: jax.Array) -> jax.Array:
    pass


@struct.dataclass
class Constraint:
  """Class specifying parameter constraints.

  `ModelParameter`s may optionally contain a `Constraint` object that specifies
  the lower/upper bounds of the parameter and a bijector that maps from the
  space of all real numbers to the interval between the lower and upper bounds.

  Attributes:
    bounds: When expressing the bounds of a single parameter (array), `bounds`
      is a tuple of arrays containing (lower, upper) bounds for the parameter.
      When expressing the bounds for a structure of parameters (such as a Flax
      parameters dict), `bounds` is a tuple of structures of the same form as
      the parameters structure, containing lower and upper bounds of
      corresponding parameters. If `bounds` is None, the parameter (or all
      parameters in the structure) are unbounded. If the tuple of bounds
      contains None, then the parameter (or corresponding parameter in the
      structure) is unbounded (from below if None is in the first element of the
      tuple, and above if None is in the second element of the tuple).
    bijector: A TFP bijector-like constraint function mapping the parameter (or
      structure of parameters) from an unconstrained space to the parameter
      domain. A value of `None` is equivalent to the identity function. A
      bijector that maps structures of parameters will typically be an instance
      of `tfb.JointMap`.
  """

  bounds: Optional[types.Bounds] = struct.field(default=None)
  bijector: Optional[tfb.Bijector] = struct.field(default=None)

  @classmethod
  def create(
      cls,
      bounds: types.Bounds,
      bijector_fn: Callable[
          [Optional[ArrayLike], Optional[ArrayLike]], tfb.Bijector
      ],
  ) -> 'Constraint':
    """Factory that builds a `Constraint` from bounds and a bijector fn.

    The constraint's bijector is created by mapping `bijector_fn` over the tree
    structure of the lower/upper bounds. The bijector will operate on the same
    nested structure as the bounds.

    Args:
      bounds: A tuple of lower and upper bounds (may be ArrayTrees matching the
        tree structure of the parameters to be constrained).
      bijector_fn: A callable that takes a lower and upper bound (Arrays) as
        args and returns a TFP bijector that maps from an unconstrained space to
        the parameter domain.

    Returns:
      constraint: A Constraint object.

    Example:

    ```python
    lower = {'a': jnp.array(0.0), 'b': jnp.array(-1.0)}
    upper = {'a': jnp.array(10.0), 'b': None}
    constraint = Constraint.create((lower, upper), bijector_fn=tfb.SoftClip)
    x = {'a': jnp.array(-0.5), 'b': jnp.array(3.0)}
    constraint.bijector(x)  # => {'a': Array(0.47), 'b': Array(3.02)}
    ```
    """
    if all(
        tree_util.treedef_is_leaf(tree_util.tree_structure(x)) for x in bounds
    ):
      bijector = bijector_fn(*bounds)
    else:
      lb, ub = bounds
      if lb is None:
        if ub is None:
          raise ValueError(
              'At least one of lower and upper bounds must be specified.'
          )
        f = lambda u: bijector_fn(None, u)
        args = (ub,)
      elif ub is None:
        f = lambda l: bijector_fn(l, None)
        args = (lb,)
      else:
        f = bijector_fn
        args = bounds
      # Use dm-tree instead of jax.tree_util to handle structures containing
      # None.
      bijector = tfb.JointMap(tree.map_structure(f, *args))
    return Constraint(bounds=bounds, bijector=bijector)


@attr.frozen
class ModelParameter:
  """Class specifying a surrogate model parameter.

  Attributes:
    name: Also used as the Flax parameter name.
    init_fn: Initializes parameter values.
    constraint: Parameter constraint.
    regularizer: Regularizes the parameter.
  """

  name: str = attr.field()
  init_fn: InitFn = attr.field()
  constraint: Optional[Constraint] = attr.field(default=None)
  regularizer: Callable[[ArrayLike], jax.Array] = attr.field(
      kw_only=True, default=lambda x: jnp.zeros([], dtype=x.dtype)
  )

  @classmethod
  def from_prior(cls,
                 prior: tfd.Distribution,
                 constraint: Optional[Constraint] = None) -> 'ModelParameter':
    """Builds a `ModelParameter` from a `tfd.Distribution`.

    If `constraint` or `constraint.bijector` is None, then the constraint
    bijector is assumed to be the prior distribution's default event space
    bijector. See
    https://www.tensorflow.org/probability/api_docs/python/tfp/distributions/Distribution#experimental_default_event_space_bijector

    Args:
      prior: Parameter prior distribution.
      constraint: The parameter constraint.

    Returns:
      model_parameter: The parameter specification with the given prior.
    """
    if constraint is None or constraint.bounds is None:
      bounds = None
    else:
      bounds = constraint.bounds
    if constraint is None or constraint.bijector is None:
      bijector = prior.experimental_default_event_space_bijector()
    else:
      bijector = constraint.bijector

    sample = lambda seed: prior.sample(seed=seed)
    init_fn = sample
    if bounds is not None:
      init_fn = lambda s: jnp.clip(sample(s), *bounds)
    return ModelParameter(
        init_fn=init_fn,
        name=prior.name,
        constraint=Constraint(bounds=bounds, bijector=bijector),
        # TODO: `jnp.copy` is used to bypass TFP bijector caching;
        # otherwise JAX tracers are leaked from JIT-ted code.
        regularizer=lambda x: -prior.log_prob(jnp.copy(x)),
    )


ModelParameterGenerator = Generator[ModelParameter, jax.Array, _D]


class ModelCoroutine(Protocol, Generic[_D]):
  """`Protocol` to avoid inheritance.

  The coroutine pattern allows the `ModelParameter` objects, and the assembly of
  parameters into the kernel and stochastic process, to be specified
  simultaneously. The `StochasticProcessModel` Flax module runs the coroutine
  to initialize Flax parameters and build stochastic process objects.

  When a `ModelCoroutine` is called, it returns a generator-iterator, which
  should be iterated to build the `ModelParameter`s and the stochastic process
  object. See the full protocol below.
  """

  def __call__(
      self,
      inputs: Optional[types.ModelInput] = None,
      **kwargs,
  ) -> ModelParameterGenerator[_D]:
    """Coroutine function to be called from `StochasticProcessModel`.

    The coroutine is implemented via an enhanced generator
    (https://peps.python.org/pep-0342/). The generator-iterator returned by this
    method corresponds to the pytype
    `Generator[YieldType, SendType, ReturnType]`. (Python also has a newer, more
    flexible `Coroutine` type declared with `async`/`await` syntax. Here, when
    we reference "coroutines," we're referring to the simpler, more restrictive
    generator-based implementation.)

    The expected protocol is to run the coroutine for two different use cases:

    1) To build the Flax model.
    2) To implement Flax model forward passes.

    During (1), a new Flax model parameter is declared with the `name` and
    `init_fn` of each `ModelParameter` yielded by the generator. The initial
    values of each Flax parameter are generated by the `init_fn` and then sent
    into the generator as the left-hand sides of the yield statements. Once all
    `ModelParameter`s are yielded, the generator raises a `StopIteration`, and
    `StopIteration.value` contains a `tfd.Distribution` representing a
    stochastic process (e.g. `tfd.GaussianProcess` or `tfd.StudentTProcess`).
    During Flax module initialization, the returned `tfd.Distribution` is
    ignored.

    During (2), for each `ModelParameter` yielded by the generator, the Flax
    module accesses the Flax parameter of the same name, regularizes it (if
    applicable), sends the value into the generator, and stores the value of the
    regularization loss in a Flax mutable variable collection. Once all
    `ModelParameter`s are yielded, the generator raises a `StopIteration`, and
    `StopIteration.value` contains a `tfd.Distribution` on the provided index
    points. The module's `__call__` method returns this distribution.

    Example:

    ```python
    # Define a coroutine for a simple Gaussian Process model with trainable
    # kernel amplitude and observation noise variance.
    def model_coroutine(inputs=None):
      amplitude_constraint = Constraint(
          bounds=(jnp.zeros([]), None), bijector=tfb.Exp())
      amplitude = yield ModelParameter(
          init_fn=jax.random.exponential,
          regularizer=lambda x: 1e-3 * x**2,
          constraint=amplitude_constraint,
          name='amplitude')
      kernel = tfpk.ExponentiatedQuadratic(amplitude=amplitude)
      observation_noise = yield ModelParameter.from_prior(
          tfd.LogNormal(0.0, 1.0, name='observation_noise'),
          constraint=Constraint(bounds=(jnp.zeros([]), None)))
      return tfd.GaussianProcess(kernel=kernel, index_points=inputs,
          observation_noise_variance=observation_noise)
    ```

    Args:
      inputs: An ArrayTree of index points or None.
      **kwargs:
    """
    pass


def _squeeze_to_event_dims(
    dist: tfd.Distribution, labels: jax.Array
) -> jax.Array:
  """Squeezes the singleton `metrics` dimension from `labels`, if applicable."""
  if len(dist.event_shape) == 1 and labels.shape == (dist.event_shape[0], 1):
    return jnp.squeeze(labels, axis=-1)
  return labels


class StochasticProcessModel(nn.Module):
  """Builds a Stochastic Process Flax module.

  The module is instantiated with a coroutine in the pattern of
  `ModelCoroutine` and represents a trainable stochastic process
  (typically a `tfd.GaussianProcess` or `tfd.StudentTProcess`.)

  The module may also be passed a `mean_fn`, which is evaluated at the input
  points and returns the mean of the stochastic process (default is a constant
  zero mean).

  Examples:

  ```python
  from jax import random

  # Simulate some observed data.
  dim = 3
  x_observed = random.uniform(random.PRNGKey(0), shape=(20, dim))
  y_observed = x_observed.sum(axis=-1)

  # Build a GP module. `coro` follows the `ModelCoroutine` protocol.
  coro = GaussianProcessARD(dimension=dim)
  gp_model = StochasticProcessModel(coroutine=coro)

  # Initialize the Flax parameters.
  init_params = gp_model.init(random.PRNGKey(1), x_observed)

  # Build a GP with `x_observed` as index points. By default, `apply` invokes
  # the Flax module's `__call__` method.
  gp, regularization_losses = gp_model.apply(
      init_params,
      x_observed,
      mutable=('losses',))

  # Run the expensive computation (often a Cholesky decomposition) necessary to
  # compute the GP posterior predictive, and return the expensive intermediates
  # as mutable state.
  _, pp_state = gp_model.apply(
      {'params': init_state['params']},
      x_observed,
      y_observed,
      method=gp_model.precompute_predictive,
      mutable=('predictive'))

  # Now, posterior predictive GPs over different sets of index points,
  # conditioned on the observed data `x_observed` and `y_observed`, can be built
  # without recomputing the Cholesky decomposition.
  x_predicted = random.uniform(random.PRNGKey(2), shape=(5, dim))
  pp_dist = gp_model.apply(
      {'params': init_state['params'], **pp_state},
      x_predicted,
      x_observed,
      y_observed,
      method=gp_model.posterior_predictive)
  ```
  """

  coroutine: ModelCoroutine

  # `None` is zero-mean.
  mean_fn: Optional[Callable[[types.ModelInput], jax.Array]] = None

  def setup(self):
    """Builds module parameters."""
    generator = self.coroutine()
    try:
      p: ModelParameter = next(generator)
      while True:
        # Declare a Flax variable with the name and initialization function from
        # the `ModelParameter`.
        param: jax.Array = self.param(p.name, p.init_fn)
        p: ModelParameter = generator.send(param)
    except StopIteration:
      # Ignore the return value from the generator since this method only builds
      # the Flax parameters.
      pass

  def __call__(self, x: types.ModelInput, **kwargs) -> _D:
    """Returns a stochastic process distribution.

    If the Flax module's `apply` method is called with `mutable=True` or
    `mutable=('losses,')` regularization losses are additionally returned.

    Args:
      x: ArrayTree of index points in the constrained space.
      **kwargs:

    Returns:
      dist: `tfd.Distribution` instance with x as index points.
    """
    gen = self.coroutine(inputs=x, **kwargs)
    if self.is_initializing() and isinstance(self.mean_fn, nn.Module):
      # If mean_fn is a module, call it so its parameters are initialized.
      _ = self.mean_fn(x)  # pylint: disable=not-callable
    try:
      p: ModelParameter = next(gen)
      while True:
        # "params" is the name that `nn.Module` gives to the collection of read-
        # only variables.
        param: jax.Array = self.get_variable('params', p.name)
        if p.regularizer:
          self.sow(  # `sow` stores a value in a collection.
              'losses',
              f'{p.name}_regularization',
              p.regularizer(param),
              reduce_fn=lambda _, b: b,
          )
        p = gen.send(param)
    except StopIteration as e:
      # After the generator is exhausted, it raises a `StopIteration` error. The
      # `StopIteration` object has a property `value` of type `_D`.
      gp = e.value
      if self.mean_fn is None:
        tfp_mean_fn = None
      else:
        # TODO: Decide on a better pattern for incorporating
        # mean_fn.
        # pylint: disable=protected-access
        def tfp_mean_fn(x_: tfpke.ContinuousAndCategoricalValues):
          return self.mean_fn(
              types.ModelInput(
                  types.PaddedArray(
                      x_.continuous,
                      fill_value=x.continuous.fill_value,
                      _original_shape=(
                          x_.continuous.shape[0],
                          x.continuous._original_shape[1],
                      ),
                      _mask=x.continuous._mask,
                      _nopadding_done=x.continuous._nopadding_done,
                  ),
                  types.PaddedArray(
                      x_.categorical,
                      fill_value=x.categorical.fill_value,
                      _original_shape=(
                          x_.categorical.shape[0],
                          x.categorical._original_shape[1],
                      ),
                      _mask=x.categorical._mask,
                      _nopadding_done=x.categorical._nopadding_done,
                  ),
              )
          )
          # pylint: enable=protected-access

      return gp.copy(mean_fn=tfp_mean_fn)

  def precompute_predictive(
      self,
      x_observed: types.ModelInput,  # Could just take ModelData
      y_observed: types.PaddedArray,
  ) -> None:
    """Builds a stochastic process regression model conditioned on observations.

    The mutable variable returned by this method as auxiliary output should be
    passed as state to `posterior_predictive`. This avoids repeated, expensive
    operations (often Cholesky decompositions) when computing the posterior
    predictive.

    Args:
      x_observed: Index points on which to condition the posterior predictive.
      y_observed: Observations on which to condition the posterior predictive.
    """
    # Call the `tfd.Distribution` object's `posterior_predictive` method. This
    # triggers an expensive computation, typically a Cholesky decomposition, and
    # returns a new `tfd.Distribution` representing the posterior predictive.
    # Expensive intermediates are stored in the `precomputed_cholesky` Flax
    # variable and returned as auxiliary output.
    kwargs = {}
    if not y_observed._nopadding_done:  # pylint: disable=protected-access
      kwargs['observations_is_missing'] = y_observed.is_missing[0]

    prior = self(x_observed)
    observations = _squeeze_to_event_dims(prior, y_observed.padded_array)
    predictive_dist = prior.posterior_predictive(
        predictive_index_points=None, observations=observations, **kwargs
    )
    # pylint: disable=protected-access
    cached_predictive_intermediates = {
        '_precomputed_divisor_matrix_cholesky': (
            predictive_dist._precomputed_divisor_matrix_cholesky
        ),
        '_precomputed_solve_on_observation': (
            predictive_dist._precomputed_solve_on_observation
        ),
    }
    # pylint: enable=protected-access
    self.sow(
        'predictive',
        'precomputed_cholesky',
        cached_predictive_intermediates,
        reduce_fn=lambda _, b: b,
    )

  def posterior_predictive(
      self,
      x_predictive: types.ModelInput,
      x_observed: types.ModelInput,
      y_observed: types.PaddedArray,
  ) -> _D:
    """Returns a posterior predictive stochastic process.

    The posterior predictive distribution over the function values at
    `x_predictive`, typically a `tfd.GaussianProcessRegressionModel` or
    `tfd.StudentTProcessRegressionModel`, is built using the mutable variable in
    `predictive/precomputed_cholesky`. This avoids repeated, expensive
    computation (often Cholesky decompositions of the kernel matrix for observed
    data). See the class docstring for how to use `precompute_predictive` in
    combination with `posterior_predictive`.

    Args:
      x_predictive: Predictive index points.
      x_observed: Index points on which to condition the posterior predictive.
      y_observed: Observations on which to condition the posterior predictive.

    Returns:
      pp_dist: The posterior predictive distribution over `x_predictive`.
    """
    if not self.has_variable('predictive', 'precomputed_cholesky'):
      raise ValueError(
          'The mutable variable returned by '
          '`precompute_predictive` must be passed into `predict`. '
          'See the class docstring for an example.'
      )
    # Access the precomputed values stored in the Flax variable, and build the
    # distribution object over the predictive index points (avoiding
    # recomputation).
    cached_intermediates = self.get_variable(
        'predictive', 'precomputed_cholesky'
    )
    kwargs = cached_intermediates
    if not y_observed._nopadding_done:  # pylint: disable=protected-access
      kwargs = kwargs | {'observations_is_missing': y_observed.is_missing[0]}

    predictive_index_points = tfpke.ContinuousAndCategoricalValues(
        continuous=x_predictive.continuous.padded_array,
        categorical=x_predictive.categorical.padded_array,
    )

    prior = self(x_observed)
    observations = _squeeze_to_event_dims(prior, y_observed.padded_array)
    return prior.posterior_predictive(
        observations=observations,
        predictive_index_points=predictive_index_points,
        **kwargs,
    )


class VectorToArrayTree(tfb.Chain):
  """Bijector that converts a vector to a dict like `params`.

  The bijector splits the vector, reshapes the splits, and then packs the splits
  into a dictionary. The bijector's `inverse` method does the reverse.

  It also has aliases `to_params` and `to_vectors` for `forward()` and
  `inverse()` methods.

  Example:

  ```python
  params = {'a': [4.0, 3.0], 'b': -2.0, 'c': [[6.0]]}
  bij = VectorToArrayTree(params)
  bij.inverse(params)  #  => [4.0, 3.0, -2.0, 6.0]
  bij([0.0, 1.0, 2.0, 3.0])  # => {'a': [0.0, 1.0], 'b': 2.0, 'c': [[3.0]]}
  ```
  """

  def __init__(
      self, arraytree: types.ParameterDict, *, validate_args: bool = False
  ):
    """Init.

    Args:
      arraytree: A nested structure of Arrays.
      validate_args: If True, does runtime validation. It may be slow.
    """
    parameters = dict(locals())
    flat, treedef = tree_util.tree_flatten(arraytree)
    restructure = tfb.Restructure(
        output_structure=tree_util.tree_unflatten(treedef, range(len(flat))))
    reshape = tfb.JointMap(
        [tfb.Reshape(event_shape_out=f.shape) for f in flat],
        validate_args=validate_args,
    )
    split = tfb.Split([x.size for x in flat], validate_args=validate_args)
    super().__init__(
        [restructure, reshape, split],
        parameters=parameters,
        name='VectorToArrayTree',
        validate_args=validate_args,
    )

  def to_params(self, vector: ArrayLike) -> types.ParameterDict:
    return self.forward(vector)

  def to_vector(self, params: types.ParameterDict) -> jax.Array:
    return self.inverse(params)


def get_constraints(
    model: StochasticProcessModel, x: Optional[Any] = None
) -> Constraint:
  """Gets the parameter constraints from a StochasticProcessModel.

  If the model contains trainable Flax variables besides those defined by the
  coroutine (for example, if `mean_fn` is a Flax module), the non-coroutine
  variables are assumed to be unconstrained (the bijector passes them through
  unmodified, and their lower/upper bounds are `None`). In this case `x` must be
  passed, so that the structure of the non-coroutine parameters dict(s) can be
  generated. `Constraint` objects for models with constrained parameters aside
  from those defined in the coroutine must be built manually.

  This method runs the coroutine, collects the parameter constraints, and
  returns a new Constraint object in which the lower/upper bounds are dicts
  (corresponding to the parameters dict) and the bijector operates on dicts. The
  object may be passed to a Vizier Optimizer to constrain the parameters for
  optimization.

  Example:

  ```python
  def model_coroutine(inputs=None):
    amplitude_constraint = Constraint(
        bounds=(jnp.array(0.1), None))
    length_scale_constraint = Constraint.create(
        bounds=(jnp.array(0.0), jnp.array(10.0)),
        bijector_fn=tfb.Sigmoid)

    amplitude = yield ModelParameter.from_prior(
        tfd.LogNormal(0.0, 1.0, name='amplitude'),
        constraint=amplitude_constraint)
    length_scale = yield ModelParameter(
        init_fn=jax.random.exponential,
        regularizer=lambda x: 1e-3 * x**2,
        constraint=length_scale_constraint,
        name='length_scale')
    kernel = tfpk.ExponentiatedQuadratic(
        amplitude=amplitude, length_scale=length_scale)
    return tfd.GaussianProcess(kernel, index_points=inputs)

  model = StochasticProcessModel(model_coroutine)
  constraint = GetConstraints(model)
  constraint.bijector
  # => tfb.JointMap({'amplitude': tfb.Exp(),
                    'length_scale': tfb.Sigmoid(0.0, 10.0)})

  constraint.bounds
  # => ({'amplitude': jnp.array(0.1), 'length_scale': jnp.array(0.0)},
  #     {'amplitude': None, 'length_scale': jnp.array(10.0)})
  ```

  Args:
    model: A `StochasticProcessModel` instance.
    x: An input that can be passed to `model.lazy_init`. `x` must be of the same
      structure as the model inputs and may contain arrays or `ShapeDtypeStruct`
      instances (see flax.linen.Module.lazy_init docs). If `model` contains Flax
      variables aside from those defined by `model.coroutine` (e.g. in a
      trainable `mean_fn`) then this arg is required.

  Returns:
    constraint: A `Constraint` instance expressing constraints on the parameters
      specified by `coroutine`.
  """

  # Run the coroutine to extract constraints for the model parameters defined in
  # the coroutine.
  gen = model.coroutine()
  k = jax.random.PRNGKey(0)
  lower = {}
  upper = {}
  bijectors = {}
  try:
    p = next(gen)
    while True:
      v = p.init_fn(k)
      if p.constraint is None or p.constraint.bounds is None:
        lower[p.name] = None
        upper[p.name] = None
      else:
        lower[p.name] = p.constraint.bounds[0]
        upper[p.name] = p.constraint.bounds[1]
      if p.constraint is None or p.constraint.bijector is None:
        bijectors[p.name] = tfb.Identity()
      else:
        bijectors[p.name] = p.constraint.bijector
      p = gen.send(v)
  except StopIteration:
    pass

  # `tfb.JointMap` applies a structure of bijectors to a parallel structure of
  # inputs. Define a `JointMap` bijector that maps an unconstrained parameters
  # dict to a constrained parameters dict with a dict of bijectors (all dicts
  # are keyed by parameter names).
  bijector = tfb.JointMap(bijectors=bijectors)
  if x is not None:
    # Get the parameters dict keys, if any, that do not come from the coroutine
    # (e.g. `mean_fn` parameters).
    params = model.lazy_init(jax.random.PRNGKey(0), x)['params']
    non_coroutine_keys = set(params.keys()) - set(bijectors.keys())

    # Define a bijector that ignores (applies an identity transformation to)
    # non-coroutine parameters.
    if non_coroutine_keys:
      logging.info(
          (
              'Defining a constraint object that ignores the following'
              'non-coroutine parameters: %s'
          ),
          non_coroutine_keys,
      )

      def _wrap_bijector_method_to_ignore_non_coro(f):
        """Wrap bijector methods to pass non-coroutine params through."""

        def _f(p):
          p_ = p.copy()
          non_coroutine_params = {k: p_.pop(k) for k in non_coroutine_keys}
          y = f(p_)
          y.update(non_coroutine_params)
          return y

        return _f

      def _bijector_fldj_with_non_coro(p):
        """Non-coroutine params do not affect the FLDJ."""
        p_ = {k: v for k, v in p.items() if k not in non_coroutine_keys}
        return bijector.forward_log_det_jacobian(p_)

      bijector_forward_min_event_ndims = bijector.forward_min_event_ndims.copy()
      # Populate `lower` and `upper` bounds dicts with `None` values for entries
      # corresponding to non-coroutine params.
      for k in non_coroutine_keys:
        lower[k] = tree.map_structure(lambda _: None, params[k])
        upper[k] = tree.map_structure(lambda _: None, params[k])
        bijector_forward_min_event_ndims[k] = tree.map_structure(
            lambda _: 0, params[k]
        )
      bijector = tfb.Inline(
          forward_fn=_wrap_bijector_method_to_ignore_non_coro(bijector.forward),
          inverse_fn=_wrap_bijector_method_to_ignore_non_coro(bijector.inverse),
          forward_log_det_jacobian_fn=_bijector_fldj_with_non_coro,
          forward_min_event_ndims=bijector_forward_min_event_ndims,
      )
    else:
      # If the model doesn't have params aside from those defined by the
      # coroutine, its params should have the same structure as `bijectors`
      # (this assertion failing indicates a bug).
      try:
        tree.assert_same_structure(params, bijectors)
      except ValueError as exc:
        raise ValueError(
            '`params` and `bijectors` should have the same nested structure.'
            f'Saw: `params={params}` and `bijectors={bijectors}`'
        ) from exc

  return Constraint((lower, upper), bijector=bijector)


class PrecomputedPredictive(eqx.Module):
  """Precomputed model for prediction.

  Conceptually, this module corresponds to the posterior distribution obtained
  by updating `prior` with `observed_data`.

  Also see `StochasticProcessWithCoroutine.precompute_predictive`.

  Attributes:
    prior: Defines the prior distribution.
    observed_data: Observations with batch shape [B].
  """

  prior: 'StochasticProcessWithCoroutine'
  observed_data: types.ModelData
  precomputed_divisor_matrix_cholesky: jax.Array
  precomputed_solve_on_observation: jax.Array

  @property
  def _posterior_kwargs(self):
    if self.observed_data.labels.shape[-1] > 1:
      # If there is more than one metric, assume we are using a MultiTask GP, in
      # which case `observations_is_missing` should be a 2D array instead of a
      # vector.
      # TODO: Make `_mask` public on PaddedArray.
      observations_is_missing = ~self.observed_data.labels._mask  # pylint: disable=protected-access
    else:
      # If there is one metric, then squeeze out the `num_metrics` dimension
      # (which has size 1).
      observations_is_missing = self.observed_data.labels.is_missing[0]
    return dict(
        _precomputed_divisor_matrix_cholesky=self.precomputed_divisor_matrix_cholesky,
        _precomputed_solve_on_observation=self.precomputed_solve_on_observation,
        observations=self.observed_data.labels.padded_array,
        observations_is_missing=observations_is_missing,
    )

  def predict(self, x_predictive: types.ModelInput) -> tfd.Distribution:
    """Returns the posterior distribution on index points `xs`.

    Args:
      x_predictive: Array(Tree) of batch shape [B].

    Returns:
      Distribution with sample shape [B].
    """
    return self._predict(x_predictive)

  def _predict(
      self,
      x_predictive: types.ModelInput,
      expand_batch_dim: bool = False,
  ) -> tfd.Distribution:
    """Returns the posterior distribution on index points `xs`.

    Args:
      x_predictive: Array(Tree) of batch shape [B].
      expand_batch_dim: If True, expand the dimensions of `x_predictive` so that
        it broadcasts with the distribution's batch shape. This is useful for
        predictions on a batch of predictive points using an ensemble GP.

    Returns:
      Distribution with sample shape [B].
    """
    x_pred_tfp = tfpke.ContinuousAndCategoricalValues(
        continuous=x_predictive.continuous.padded_array,
        categorical=x_predictive.categorical.padded_array,
    )
    if expand_batch_dim:
      x_pred_tfp = tfpke.ContinuousAndCategoricalValues(
          continuous=x_pred_tfp.continuous[:, jnp.newaxis, ...],
          categorical=x_pred_tfp.categorical[:, jnp.newaxis, ...],
      )
    prior_gp = self.prior(self.observed_data.features)
    kwargs = self._posterior_kwargs.copy()
    kwargs['observations'] = _squeeze_to_event_dims(
        prior_gp, kwargs['observations']
    )
    return prior_gp.posterior_predictive(
        predictive_index_points=x_pred_tfp, **kwargs
    )


class UniformEnsemblePredictive(eqx.Module):
  """Uniform ensemble of predictive models.

  Ensembles the `predictives` with equal weights.
  """

  predictives: PrecomputedPredictive

  def predict(self, xs: types.ModelInput) -> tfd.Distribution:
    return self.predict_with_aux(xs)[0]

  def predict_with_aux(
      self, xs: types.ModelInput
  ) -> tuple[tfd.Distribution, chex.ArrayTree]:
    # If `xs` has a batch dimension and the predictive distribution is
    # ensembled, we expand xs' batch shape so as not to collide with the
    # ensemble dimensions. (`vmap` cannot currently be used on functions
    # returning TFP distributions).
    has_batched_hparams = (
        self.predictives.precomputed_solve_on_observation.ndim > 1
    )
    expand_x = len(xs.continuous.shape) == 3 and has_batched_hparams
    dist = self.predictives._predict(xs, expand_batch_dim=expand_x)  # pylint: disable=protected-access
    if has_batched_hparams:
      return (
          tfd.MixtureSameFamily(
              tfd.Categorical(logits=jnp.zeros(dist.batch_shape[-1])), dist
          ),
          {
              'components_mean': dist.mean().T,
              'components_stddev': dist.stddev().T,
          },
      )
    return dist, {}


def _initialize_params(
    coroutine: ModelCoroutine, rng: jax.Array
) -> chex.ArrayTree:
  """Randomly initializes a coroutine's parameters."""
  gen = coroutine()
  params = {}
  try:
    p: ModelParameter = next(gen)
    while True:
      # Declare a Flax variable with the name and initialization function from
      # the `ModelParameter`.
      rng, init_rng = jax.random.split(rng)
      param = p.init_fn(init_rng)
      params[p.name] = param
      p: ModelParameter = gen.send(param)
  except StopIteration:
    return params


class StochasticProcessWithCoroutine(eqx.Module):
  """Fully parameterized stochastic process model."""

  coroutine: ModelCoroutine = eqx.field(static=True)
  params: chex.ArrayTree

  @classmethod
  def initialize(cls, coroutine: ModelCoroutine, *, rng: jax.Array):
    """Builds module parameters."""
    # return cls(coroutine, params=_initialize_params(coroutine, rng))
    # TODO: Put back the line above. This code currently uses
    # flax's init routine to exactly match the setup so we can replicate
    # the trajectories from pre-refactoring.
    return cls(
        coroutine, StochasticProcessModel(coroutine).init(rng, None)['params']
    )

  def call_with_aux(
      self, x: types.ModelInput, /
  ) -> tuple[tfd.Distribution, chex.ArrayTree]:
    """Returns the joint distribution and auxiliary information.

    Args:
      x:

    Returns:
      (dist, aux)
      dist: Distribution
      aux: A dict where all the leaf nodes in the subtree aux[`losses`] contain
        regularization losses.
    """
    gen = self.coroutine(x)
    params = self.params
    params_loss = dict()
    try:
      p: ModelParameter = next(gen)
      while True:
        if p.regularizer:
          params_loss[p.name] = p.regularizer(params[p.name])
        # "params" is the name that `nn.Module` gives to the collection of read-
        # only variables.
        p = gen.send(params[p.name])
    except StopIteration as e:
      # After the generator is exhausted, it raises a `StopIteration` error. The
      # `StopIteration` object has a property `value` of type `_D`.
      return e.value, {'losses': params_loss}

  def __call__(self, x: types.ModelInput) -> tfd.Distribution:
    return self.call_with_aux(x)[0]

  def loss_with_aux(
      self,
      data: types.ModelData,
      seed: Optional[jax.Array] = None,
  ) -> tuple[jax.Array, chex.ArrayTree]:
    dist, aux = self.call_with_aux(data.features)

    # If `seed` is provided, that implies `dist.log_prob` is a stochastic
    # approximation that requires a random seed.
    log_prob_kwargs = {}
    if seed is not None:
      log_prob_kwargs['key'] = seed

    labels = _squeeze_to_event_dims(dist, data.labels.padded_array)
    # TODO: Enable `is_missing` for MTGP.
    if isinstance(dist, tfde.MultiTaskGaussianProcess):
      logging.warning(
          'Using a multitask GP; note that padding/masking is not yet supported'
          'in `log_prob`.'
      )
      nll_data = -dist.log_prob(labels, **log_prob_kwargs)
    else:
      nll_data = -dist.log_prob(
          labels, is_missing=data.labels.is_missing[0], **log_prob_kwargs
      )
    loss = nll_data + jax.tree_util.tree_reduce(jnp.add, aux['losses'])
    return loss, aux

  def precompute_predictive(
      self, data: types.ModelData
  ) -> PrecomputedPredictive:
    jax.monitoring.record_event(
        '/vizier/jax/coroutine_with_params/precompute_predictive/traced'
    )
    if jnp.size(data.labels.padded_array) == 0:
      # TFP `retrying_cholesky` does not handle empty observations.
      prior = self(data.features).copy(cholesky_fn=None)
    else:
      prior = self(data.features)

    observations = _squeeze_to_event_dims(prior, data.labels.padded_array)
    if isinstance(prior, tfde.MultiTaskGaussianProcess):
      # TODO: Make `_mask` public on PaddedArray.
      observations_is_missing = ~data.labels._mask  # pylint: disable=protected-access
    else:
      observations_is_missing = data.labels.is_missing[0]
    predictive = prior.posterior_predictive(
        predictive_index_points=None,
        observations=observations,
        observations_is_missing=observations_is_missing,
    )
    # pylint: disable=protected-access
    return PrecomputedPredictive(
        self,
        data,
        predictive._precomputed_divisor_matrix_cholesky,
        predictive._precomputed_solve_on_observation,
    )


class CoroutineWithData(eqx.Module):
  """Utility module for training (ARD).

  Has setup() and loss_with_aux() that can be readily `eqx.filter_jit`ed
  """

  coroutine: ModelCoroutine = eqx.field(static=True)
  data: types.ModelData

  def constraints(self):
    return get_constraints(StochasticProcessModel(self.coroutine))

  def setup(self, rng: jax.Array):
    jax.monitoring.record_event('/vizier/jax/coroutine_with_data/setup/traced')
    return StochasticProcessWithCoroutine.initialize(
        self.coroutine, rng=rng
    ).params

  def loss_with_aux(
      self, params, seed: Optional[jax.Array] = None
  ) -> tuple[jax.Array, chex.ArrayTree]:
    jax.monitoring.record_event(
        '/vizier/jax/coroutine_with_data/loss_with_aux/traced'
    )
    return StochasticProcessWithCoroutine(self.coroutine, params).loss_with_aux(
        self.data, seed=seed
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for exploration_score_utils."""

from typing import Tuple

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.analyzers import exploration_score_utils

from absl.testing import absltest
from absl.testing import parameterized


def _generate_min_and_max_ent_studies() -> (
    Tuple[vz.ProblemAndTrials, vz.ProblemAndTrials]
):
  """Generates two studies with zero and large parameter entropies."""
  space = vz.SearchSpace()
  root = space.root
  root.add_float_param('continuous', -5.0, 5.0)
  root.add_int_param('integer', -5, 5)
  root.add_categorical_param(
      'categorical', [str(v) for v in np.linspace(-5, 5, 11)]
  )
  root.add_discrete_param('discrete', list(np.linspace(-5, 5, 11)))
  problem = vz.ProblemStatement(
      search_space=space,
      metric_information=[
          vz.MetricInformation('x1', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
      ],
  )
  max_ent_trials = []
  min_ent_trials = []
  values = list(np.linspace(-5, 5, 11)) * 10
  for idx, value in enumerate(values):
    # Generates trials with large marginal parameter entropies by looping
    # through the feasible values of each parameter.
    max_ent_trials.append(
        vz.Trial(
            id=idx + 1,
            parameters={
                'continuous': vz.ParameterValue(value),
                'integer': vz.ParameterValue(int(value)),
                'categorical': vz.ParameterValue(str(value)),
                'discrete': vz.ParameterValue(value),
            },
        )
    )
    # Generates trials with zero marginal parameter entropies by setting every
    # parameter to the same value.
    min_ent_trials.append(
        vz.Trial(
            id=idx + 1,
            parameters={
                'continuous': vz.ParameterValue(values[55]),
                'integer': vz.ParameterValue(int(values[17])),
                'categorical': vz.ParameterValue(str(values[96])),
                'discrete': vz.ParameterValue(values[34]),
            },
        )
    )
  return vz.ProblemAndTrials(
      problem=problem, trials=min_ent_trials
  ), vz.ProblemAndTrials(problem=problem, trials=max_ent_trials)


class ExplorationScoreUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Continuous', vz.ParameterType.DOUBLE),
      ('Discrete', vz.ParameterType.DISCRETE),
      ('Integer', vz.ParameterType.INTEGER),
      ('Categorical', vz.ParameterType.CATEGORICAL),
  )
  def test_compute_parameter_entropy(self, parameter_type):
    def _cast(v):
      if parameter_type == vz.ParameterType.INTEGER:
        return int(v)
      elif parameter_type == vz.ParameterType.CATEGORICAL:
        return str(v)
      else:
        return v

    feasible_values = [_cast(v) for v in np.linspace(-5, 5, 11)]
    if parameter_type in [vz.ParameterType.DOUBLE, vz.ParameterType.INTEGER]:
      parameter_config = vz.ParameterConfig.factory(
          name='param',
          bounds=(-5, 5)
          if parameter_type == vz.ParameterType.INTEGER
          else (-5.0, 5.0),
      )
    else:
      parameter_config = vz.ParameterConfig.factory(
          name='param', feasible_values=feasible_values
      )
    parameter_values = [vz.ParameterValue(value=v) for v in feasible_values]
    max_ent_parameter_values = parameter_values * 10
    min_ent_parameter_values = [parameter_values[5]] * len(
        max_ent_parameter_values
    )

    max_ent = exploration_score_utils.compute_parameter_entropy(
        parameter_config=parameter_config,
        parameter_values=max_ent_parameter_values,
    )
    min_ent = exploration_score_utils.compute_parameter_entropy(
        parameter_config=parameter_config,
        parameter_values=min_ent_parameter_values,
    )
    self.assertAlmostEqual(min_ent, 0)
    if parameter_type != vz.ParameterType.DOUBLE:
      self.assertAlmostEqual(max_ent, -np.log(1 / 11))
    else:
      # For continuous parameters the entropy depends on the histogram, whose
      # number of bins is chosen based on data. There are 11 unique values with
      # equal counts, and we simply assert that the entropy is greater than
      # the entropy of two bins with equal counts.
      self.assertGreater(max_ent, -np.log(0.5))

  def test_compute_average_marginal_parameter_entropy(self):
    min_ent_study, max_ent_study = _generate_min_and_max_ent_studies()
    max_ent = (
        exploration_score_utils.compute_average_marginal_parameter_entropy(
            {'spec_gen': {'hash_': {0: max_ent_study}}}
        )
    )
    min_ent = (
        exploration_score_utils.compute_average_marginal_parameter_entropy(
            {'spec_gen': {'hash_': {0: min_ent_study}}}
        )
    )
    self.assertAlmostEqual(min_ent, 0)
    # The study has four parameters: continuous, integer, categorical and
    # discrete. The entropy for all parameters except the continuous are
    # exactly -np.log(1/11) since every parameter has 11 unique values with
    # equal counts. The entropy of the continuous parameter depends on the
    # histogram, whose number of bins is chosen based on data, and we simply
    # lower bound it by the entropy of two bins with equal counts.
    self.assertGreater(
        max_ent, 0.75 * (-np.log(1 / 11)) + 0.25 * (-np.log(0.5))
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gp_bandit."""

import numpy as np
from vizier.pyvizier.converters import padding

from absl.testing import absltest
from absl.testing import parameterized


class PaddingTest(parameterized.TestCase):

  def test_padding(self):
    features = np.random.randn(13, 7)
    labels = np.random.randn(13, 3)

    schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.MULTIPLES_OF_10,
        num_features=padding.PaddingType.POWERS_OF_2,
        num_metrics=padding.PaddingType.POWERS_OF_2,
    )

    padded = schedule.pad_features(features)

    self.assertSequenceEqual(padded.padded_array.shape, (20, 8))
    self.assertSequenceEqual(padded.unpad().shape, features.shape)

    padded = schedule.pad_labels(labels)
    self.assertSequenceEqual(padded.padded_array.shape, (20, 4))
    self.assertSequenceEqual(padded.unpad().shape, labels.shape)

  def test_nopadding(self):
    features = np.random.randn(13, 7)
    labels = np.random.randn(13, 3)

    schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.NONE,
        num_features=padding.PaddingType.NONE,
        num_metrics=padding.PaddingType.NONE,
    )

    padded = schedule.pad_features(features)
    self.assertSequenceEqual(padded.padded_array.shape, features.shape)

    padded = schedule.pad_features(labels)
    self.assertSequenceEqual(padded.padded_array.shape, labels.shape)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for scheduled_gp_bandit."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import scheduled_gp_bandit
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies
from absl.testing import absltest


class ScheduledGpBanditTest(absltest.TestCase):

  def test_schedule_designer(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    designer = scheduled_gp_bandit.VizierScheduledGPBandit(
        problem,
        max_num_trials=5,
        init_ucb_coef=4.0,
        final_ucb_coef=1.0,
        decay_rate=1.2,
    )
    self.assertEqual(designer._compute_ucb_coefficient(), 4.0)
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=5,
            batch_size=1,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(designer),
        5,
    )
    self.assertEqual(designer._compute_ucb_coefficient(), 1.0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier import pyvizier
from vizier._src.pyvizier.multimetric import safety
from absl.testing import absltest


class SafeAcquisitionTest(absltest.TestCase):

  def testSafeMeasurements(self):
    info1 = pyvizier.MetricInformation(
        name='safety1',
        goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE,
        safety_threshold=0.1)
    info2 = pyvizier.MetricInformation(
        name='safety2',
        goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
        safety_threshold=0.5)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1, info2]))
    safe_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=3.1),
        'safety2': pyvizier.Metric(value=0.3)
    })
    unsafe1_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=-3.1),
        'safety2': pyvizier.Metric(value=0.4)
    })
    unsafe2_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=3.1),
        'safety2': pyvizier.Metric(value=0.7)
    })
    no_safety_measurement = pyvizier.Measurement(
        metrics={'safety1': pyvizier.Metric(value=3.1)})

    self.assertListEqual(
        checker.are_measurements_safe([
            safe_measurement, unsafe1_measurement, unsafe2_measurement,
            no_safety_measurement
        ]), [True, False, False, True])

  def testSafeTrials(self):
    info1 = pyvizier.MetricInformation(
        name='obj', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
    info2 = pyvizier.MetricInformation(
        name='safety',
        goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
        safety_threshold=0.5)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1, info2]))

    safe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.3
        }))
    unsafe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.7
        }))
    no_safety_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={'obj': 0.2}))

    # Empty Trial is assumed to be safe.
    self.assertListEqual(
        checker.are_trials_safe(
            [safe_trial, unsafe_trial, no_safety_trial,
             pyvizier.Trial()]), [True, False, True, True])

  def testNoSafety(self):
    info1 = pyvizier.MetricInformation(
        name='obj', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1]))

    safe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.3
        }))
    unsafe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.7
        }))

    # With no safety metrics, all Trials are safe..
    self.assertListEqual(
        checker.are_trials_safe([safe_trial, unsafe_trial,
                                 pyvizier.Trial()]), [True, True, True])

  def testWarpUnsafeTrials(self):
    info1 = pyvizier.MetricInformation(
        name='obj', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
    )
    info2 = pyvizier.MetricInformation(
        name='safety',
        goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
        safety_threshold=0.5,
    )
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1, info2]))

    safe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(
            metrics={'obj': 0.2, 'safety': 0.3}
        )
    )
    unsafe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(
            metrics={'obj': 0.2, 'safety': 0.7, 'extra': 0.3}
        )
    )
    no_safety_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={'obj': 0.2})
    )

    # Empty safety measurement is assumed to be safe.
    warped_trials = checker.warp_unsafe_trials(
        [safe_trial, unsafe_trial, no_safety_trial]
    )

    self.assertEqual(warped_trials[0], safe_trial)
    self.assertEqual(warped_trials[2], no_safety_trial)

    # Unsafe Trial is warped.
    self.assertEqual(
        warped_trials[1].final_measurement.metrics['obj'].value, -np.inf
    )
    # Extra metrics should be untouched.
    self.assertEqual(
        warped_trials[1].final_measurement.metrics['extra'].value, 0.3
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""A Vizier Ray Searcher."""

import datetime
import json
from typing import Dict, Optional
import uuid

from ray import tune
from ray.tune import search
from vizier._src.raytune import converters
from vizier.service import clients
from vizier.service import pyvizier as svz


class VizierSearch(search.Searcher):
  """An OSS Vizier Searcher for Ray."""

  def __init__(
      self,
      study_id: Optional[str] = None,
      problem: Optional[svz.StudyConfig] = None,
      algorithm: Optional[str] = 'GAUSSIAN_PROCESS_BANDIT',
      **kwargs,
  ):
    """Initialize a Searcher via ProblemStatement.

    To initialize VizierSearch via set_search_properties, do not set problem.

    Args:
      study_id: The study id in the Vizier service.
      problem: The study config to optimize over. `problem.algorithm` is
        overwritten by `algorithm`.
      algorithm: The Vizier algorithm to use. Overrides the algorithm in
        problem.
      **kwargs:
    """
    super().__init__(**kwargs)

    if study_id:
      self.study_id = study_id
    else:
      self.study_id = f'ray_vizier_{uuid.uuid1()}'

    self.algorithm = algorithm

    # Mapping from Ray trial id to Vizier Trial client.
    self._active_trials: Dict[str, clients.Trial] = {}

    # The name of the metric being optimized, for single objective studies.
    self._metric = None

    # Vizier service client.
    self.study_client: Optional[clients.Study] = None
    if problem:
      if not problem.is_single_objective:
        raise ValueError(
            f'Only single objective studies are supported: {problem}'
        )
      # We can't store StudyConfig since it contains a proto, and it's not
      # pickleable, so we store problem statement instead.
      # TODO: store a proto string instead.
      self._problem = problem.to_problem()
      self._metric = problem.metric_information.item().name
      if self.algorithm is None:
        self.algorithm = problem.algorithm

  def set_search_properties(
      self, metric: Optional[str], mode: Optional[str], config: Dict, **spec  # pylint: disable=g-bare-generic
  ) -> bool:
    """Pass search properties to searcher.

    This method acts as an alternative to instantiating search algorithms
    with their own specific search spaces. Instead they can accept a
    Tune config through this method. A searcher should return ``True``
    if setting the config was successful, or ``False`` if it was
    unsuccessful, e.g. when the search space has already been set.

    Args:
        metric: Metric to optimize
        mode: One of ["min", "max"]. Direction to optimize.
        config: Tune config dict.
        **spec: Any kwargs for forward compatiblity. Info like
          Experiment.PUBLIC_KEYS is provided through here.

    Returns:
      True on success, False on failure.
    """
    if self.study_client:
      # The study is already configured.
      return False

    if mode not in ['min', 'max']:
      raise ValueError("'mode' must be one of ['min', 'max']")

    self._metric = metric or tune.result.DEFAULT_METRIC

    search_space = converters.SearchSpaceConverter.to_vizier(config)
    vizier_goal = (
        svz.ObjectiveMetricGoal.MAXIMIZE
        if mode == 'max'
        else svz.ObjectiveMetricGoal.MINIMIZE
    )
    study_config = svz.StudyConfig(
        search_space=search_space,
        algorithm=self.algorithm,
        metric_information=[
            svz.MetricInformation(self._metric, goal=vizier_goal)
        ],
    )
    self.study_client = clients.Study.from_study_config(
        study_config, owner='raytune', study_id=self.study_id
    )
    return True

  def on_trial_result(self, trial_id: str, result: Dict) -> None:  # pylint: disable=g-bare-generic
    if trial_id not in self._active_trials:
      raise RuntimeError(f'No active trial for {trial_id}')
    if self.study_client is None:
      raise RuntimeError(
          'VizierSearch not initialized! Set a search space first.'
      )
    trial_client = self._active_trials[trial_id]
    elapsed_secs = (
        datetime.datetime.now().astimezone()
        - trial_client.materialize().creation_time
    )
    trial_client.add_measurement(
        svz.Measurement(
            {k: v for k, v in result.items() if isinstance(v, float)},
            elapsed_secs=elapsed_secs.total_seconds(),
        )
    )

  def on_trial_complete(
      self, trial_id: str, result: Optional[Dict] = None, error: bool = False  # pylint: disable=g-bare-generic
  ) -> None:
    if trial_id not in self._active_trials:
      raise RuntimeError(f'No active trial for {trial_id}')
    if self.study_client is None:
      raise RuntimeError(
          'VizierSearch not initialized! Set a search space first.'
      )
    trial_client = self._active_trials[trial_id]

    if error:
      # Mark the trial as infeasible.
      trial_client.complete(
          infeasible_reason=f'Trial {trial_id} failed: {result}'
      )
    else:
      measurement = None
      if result:
        elapsed_secs = (
            datetime.datetime.now().astimezone()
            - trial_client.materialize().creation_time
        )
        measurement = svz.Measurement(
            {k: v for k, v in result.items() if isinstance(v, float)},
            elapsed_secs=elapsed_secs.total_seconds(),
        )
      trial_client.complete(measurement=measurement)
    self._active_trials.pop(trial_id)

  def suggest(self, trial_id):
    if self.study_client is None:
      study_config = svz.StudyConfig.from_problem(self._problem)
      study_config.algorithm = self.algorithm
      self.study_client = clients.Study.from_study_config(
          study_config, owner='raytune', study_id=self.study_id
      )
    suggestions = self.study_client.suggest(count=1, client_id=trial_id)
    if not suggestions:
      return search.Searcher.FINISHED

    self._active_trials[trial_id] = suggestions[0]
    return self._active_trials[trial_id].parameters

  # TODO: Test save and restore.
  def save(self, checkpoint_path):
    # We assume that the Vizier service continues running, so the only
    # information needed to restore this searcher is the mapping from the Ray
    # to Vizier trial ids. All other information can become stale and is best
    # restored from the Vizier service in restore().
    ray_to_vizier_trial_ids = {}
    for trial_id, trial_client in self._active_trials.items():
      ray_to_vizier_trial_ids[trial_id] = trial_client.id
    with open(checkpoint_path, 'w') as f:
      json.dump(
          {
              'study_id': self.study_id,
              'ray_to_vizier_trial_ids': ray_to_vizier_trial_ids,
          },
          f,
      )

  def restore(self, checkpoint_path):
    with open(checkpoint_path, 'r') as f:
      obj = json.load(f)

    self.study_id = obj['study_id']
    self.study_client = clients.Study.from_owner_and_id(
        'raytune', self.study_id
    )
    self._metric = (
        self.study_client.materialize_study_config().metric_information.item()
    )
    self._active_trials = {}
    for ray_id, vizier_trial_id in obj['ray_to_vizier_trial_ids'].items():
      self._active_trials[ray_id] = self.study_client.get_trial(vizier_trial_id)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Mappers between different feature formats."""

import jax
from jax import numpy as jnp
import numpy as np
from vizier._src.jax import types
from vizier.pyvizier.converters import core


class ContinuousCategoricalFeatureMapper:
  """Maps features from TrialToArrayConverter to ContinuousAndCategoricalValues.

  The class allows to separate the continuous features from the categorical
  features and organize them in an ContinuousAndCategoricalValues namedtuple.

  Naming convension:
    parameter - Vizier parameter.
    feature - Numpy array representing a converted Vizier parameter.

  Notation for dimensions:
    C - continuous parameters count
    P - categorical parameters count
    F - total categorial dimensions (i.e. the total feasible values count, which
      associated with the onehot-encoding features total count)
    B - number of trials
  """

  def __init__(self, converter: core.TrialToArrayConverter):
    """Initiate helper objects to split features by type."""
    # Number of categorical parameters.
    self.n_categorical_params = 0
    # The continuous feature indices of output_spec, which also correspond to
    # the columns indices of continuous parameters in features.
    self._continuous_indices = []
    # The categorical feature indices.
    self._categorical_indices = []
    # The dimensions of categorical parameters.
    categorical_dims = []
    ind = 0
    # Extract the continuous and categorical feature indices.
    for spec in converter.output_specs:
      if spec.type == core.NumpyArraySpecType.CONTINUOUS:
        self._continuous_indices.append(ind)
        ind += 1
      elif spec.type == core.NumpyArraySpecType.ONEHOT_EMBEDDING:
        self._categorical_indices.extend(range(ind, ind + spec.num_dimensions))
        ind += spec.num_dimensions
        self.n_categorical_params += 1
        categorical_dims.append(spec.num_dimensions)
      elif spec.type == core.NumpyArraySpecType.DISCRETE:
        # There shouldn't be DISCRETE spec type as onehot_embed=True is the
        # default in TrialToArrayConverter, and DISCRETE parameters are either
        # converted to CONTINUOUS or ONEHOT_EMBEDDING.
        raise ValueError("DISCRETE spec type shouldn't exists.")
      else:
        raise ValueError("Unexpected spec type %s!" % spec.type)
    # Create shift array containing the cumulative number of categorical
    # dimensions until each categorical parameter (itself not included). One-hot
    # active bit (index) is mapped to integer value, by subtracting 'shift'.
    # For example, if categorical_dims=[3, 2], then cumsum(pad(..)) will
    # generate [0, 3, 5], so 'shift' will be [0, 3].
    self.shift = jnp.cumsum(jnp.pad(jnp.array(categorical_dims), (1, 0)))[
        0:-1
    ]  # (P,)
    self.categorical_dims = categorical_dims
    self.converter = converter

  def map(self, features: types.Array) -> types.ContinuousAndCategoricalArray:
    """Split features by 'continuous' and 'categorical'.

    In addition to splitting the method converts the one-hot encoding
    categorical features to integer indices.

    For example, if the search space is:
      - categorical parameter C1 with values ['a', 'b', 'c].
      - continuous parameter F1.
      - categorical parameter C2 with values ['x', 'y'].

    Then for input parameters C1='b', F1=0.23, C2='y' and associated features:
    [[0, 1, 0, 0.23, 0, 1]] the result would be:
    ContinuousAndCategoricalValues(
        continuous=[[0.23]]
        categorical=[[1, 1]])

    Arguments:
      features: Numpy array (n_trials, n_features) which is the output of
        'to_features' method.

    Returns:
      namedtuple with continuous features and categorical integer indices.
    """
    # Split features to continuous and categorical.
    batch_shape = features.shape[:-1]
    continuous_features = features[..., self._continuous_indices]  # (B,C)
    # Assign empty array as the default value.
    categorical_index_features = jnp.zeros(
        batch_shape + (0,), dtype=types.INT_DTYPE
    )
    if self.n_categorical_params > 0:
      categorical_features = features[..., self._categorical_indices]  # (B,F)
      # Find the non-zero column indices associated with categorical parameters.
      # For example (cont. from above), with the input of [0,1,0, 0.23, 0,1]
      # 'categorical_features' is [[0,1,0,0,1]] and 'nonzero_indices' is [1,4]
      nonzero_size = np.prod(batch_shape) * self.n_categorical_params
      # nonzero_size: (P*B,)
      nonzero_indices = jnp.nonzero(categorical_features, size=nonzero_size)[1]
      # Reshape the non-zero indices to align with the no. of categorical params
      # and shift by the cardinality of each parameter to compute the integer
      # category value. For example (cont. from above), [[1,4]] - [0,3] = [1, 1]
      categorical_index_features = (
          jnp.reshape(
              nonzero_indices, batch_shape + (self.n_categorical_params,)
          )  # (B,P)
          - self.shift  # (P,)
      ).astype(
          types.INT_DTYPE
      )  # (B,P)
    return types.ContinuousAndCategoricalArray(
        continuous=continuous_features, categorical=categorical_index_features
    )

  def unmap(self, features: types.ContinuousAndCategoricalArray) -> jax.Array:
    """Convert back from ContinuousAndCategoricalArray to features."""
    if features.continuous.shape[0] != features.categorical.shape[0]:
      raise ValueError(
          "'continuous' and 'categorical' first dimension doesn't match!"
      )
    batch_shape = features.continuous.shape[:-1]
    unmapped_features = jnp.zeros(
        batch_shape
        + (sum(self.categorical_dims) + len(self._continuous_indices),)
    )
    con_ind = 0
    cat_ind = 0
    ind = 0
    # Extract the continuous and categorical feature indices.
    for spec in self.converter.output_specs:
      if spec.type == core.NumpyArraySpecType.CONTINUOUS:
        unmapped_features = unmapped_features.at[..., ind].set(
            features.continuous[..., con_ind]
        )
        con_ind += 1
        ind += 1

      elif spec.type == core.NumpyArraySpecType.ONEHOT_EMBEDDING:
        unmapped_features = unmapped_features.at[
            ..., ind : ind + spec.num_dimensions
        ].set(jnp.eye(spec.num_dimensions)[features.categorical[..., cat_ind]])
        cat_ind += 1
        ind += spec.num_dimensions
      else:
        raise ValueError("Unexpected spec type %s!" % spec.type)

    return unmapped_features
